@@ -1,21 +1,28 @@
-use super::{Process, State};
-use crate::core::data::{
-    DirEntry, FileType, Metadata, Request, RequestPayload, Response, ResponsePayload,
-    RunningProcess,
+use crate::core::{
+    data::{
+        self, DirEntry, FileType, Metadata, Request, RequestPayload, Response, ResponsePayload,
+        RunningProcess,
+    },
+    state::{Process, ServerState},
 };
 use log::*;
 use std::{
-    error::Error, net::SocketAddr, path::PathBuf, process::Stdio, sync::Arc, time::SystemTime,
+    error::Error,
+    net::SocketAddr,
+    path::{Path, PathBuf},
+    process::Stdio,
+    sync::Arc,
+    time::SystemTime,
 };
 use tokio::{
-    io::{self, AsyncReadExt, AsyncWriteExt},
+    io::{self, AsyncBufReadExt, AsyncWriteExt, BufReader},
     process::Command,
     sync::{mpsc, oneshot, Mutex},
 };
 use walkdir::WalkDir;
 
 pub type Reply = mpsc::Sender<Response>;
-type HState = Arc<Mutex<State>>;
+type HState = Arc<Mutex<ServerState<SocketAddr>>>;
 
 /// Processes the provided request, sending replies using the given sender
 pub(super) async fn process(
@@ -38,7 +45,12 @@ pub(super) async fn process(
             RequestPayload::FileWriteText { path, text } => file_write(path, text).await,
             RequestPayload::FileAppend { path, data } => file_append(path, data).await,
             RequestPayload::FileAppendText { path, text } => file_append(path, text).await,
-            RequestPayload::DirRead { path, all } => dir_read(path, all).await,
+            RequestPayload::DirRead {
+                path,
+                depth,
+                absolute,
+                canonicalize,
+            } => dir_read(path, depth, absolute, canonicalize).await,
             RequestPayload::DirCreate { path, all } => dir_create(path, all).await,
             RequestPayload::Remove { path, force } => remove(path, force).await,
             RequestPayload::Copy { src, dst } => copy(src, dst).await,
@@ -48,7 +60,7 @@ pub(super) async fn process(
                 proc_run(tenant.to_string(), addr, state, tx, cmd, args).await
             }
             RequestPayload::ProcKill { id } => proc_kill(state, id).await,
-            RequestPayload::ProcStdin { id, data } => proc_stdin(state, id, data).await,
+            RequestPayload::ProcStdin { id, line } => proc_stdin(state, id, line).await,
             RequestPayload::ProcList {} => proc_list(state).await,
         }
     }
@@ -107,21 +119,57 @@ async fn file_append(
     Ok(ResponsePayload::Ok)
 }
 
-async fn dir_read(path: PathBuf, all: bool) -> Result<ResponsePayload, Box<dyn Error>> {
+async fn dir_read(
+    path: PathBuf,
+    depth: usize,
+    absolute: bool,
+    canonicalize: bool,
+) -> Result<ResponsePayload, Box<dyn Error>> {
+    // Canonicalize our provided path to ensure that it is exists, not a loop, and absolute
+    let root_path = tokio::fs::canonicalize(path).await?;
+
     // Traverse, but don't include root directory in entries (hence min depth 1)
-    let dir = WalkDir::new(path.as_path()).min_depth(1);
+    let dir = WalkDir::new(root_path.as_path()).min_depth(1);
 
-    // If all, will recursively traverse, otherwise just return directly from dir
-    let dir = if all { dir } else { dir.max_depth(1) };
+    // If depth > 0, will recursively traverse to specified max depth, otherwise
+    // performs infinite traversal
+    let dir = if depth > 0 { dir.max_depth(depth) } else { dir };
 
-    // TODO: Support both returning errors and successfully-traversed entries
-    // TODO: Support returning full paths instead of always relative?
-    Ok(ResponsePayload::DirEntries {
-        entries: dir
-            .into_iter()
-            .map(|e| {
-                e.map(|e| DirEntry {
-                    path: e.path().strip_prefix(path.as_path()).unwrap().to_path_buf(),
+    // Determine our entries and errors
+    let mut entries = Vec::new();
+    let mut errors = Vec::new();
+
+    for entry in dir {
+        match entry.map_err(data::Error::from) {
+            Ok(e) => {
+                // Canonicalize the path if specified, otherwise just return
+                // the path as is
+                let mut path = if canonicalize {
+                    match tokio::fs::canonicalize(e.path()).await {
+                        Ok(path) => path,
+                        Err(x) => {
+                            errors.push(data::Error::from(x));
+                            continue;
+                        }
+                    }
+                } else {
+                    e.path().to_path_buf()
+                };
+
+                // Strip the path of its prefix based if not flagged as absolute
+                if !absolute {
+                    // NOTE: In the situation where we canonicalized the path earlier,
+                    //       there is no guarantee that our root path is still the
+                    //       parent of the symlink's destination; so, in that case we MUST just
+                    //       return the path if the strip_prefix fails
+                    path = path
+                        .strip_prefix(root_path.as_path())
+                        .map(Path::to_path_buf)
+                        .unwrap_or(path);
+                };
+
+                entries.push(DirEntry {
+                    path,
                     file_type: if e.file_type().is_dir() {
                         FileType::Dir
                     } else if e.file_type().is_file() {
@@ -130,10 +178,13 @@ async fn dir_read(path: PathBuf, all: bool) -> Result<ResponsePayload, Box<dyn E
                         FileType::SymLink
                     },
                     depth: e.depth(),
-                })
-            })
-            .collect::<Result<Vec<DirEntry>, walkdir::Error>>()?,
-    })
+                });
+            }
+            Err(x) => errors.push(x),
+        }
+    }
+
+    Ok(ResponsePayload::DirEntries { entries, errors })
 }
 
 async fn dir_create(path: PathBuf, all: bool) -> Result<ResponsePayload, Box<dyn Error>> {
@@ -260,16 +311,15 @@ async fn proc_run(
     // Spawn a task that sends stdout as a response
     let tx_2 = tx.clone();
     let tenant_2 = tenant.clone();
-    let mut stdout = child.stdout.take().unwrap();
+    let mut stdout = BufReader::new(child.stdout.take().unwrap()).lines();
     let stdout_task = tokio::spawn(async move {
         loop {
-            let mut data = Vec::new();
-            match stdout.read_to_end(&mut data).await {
-                Ok(n) if n > 0 => {
+            match stdout.next_line().await {
+                Ok(Some(line)) => {
                     let res = Response::new(
                         tenant_2.as_str(),
                         None,
-                        ResponsePayload::ProcStdout { id, data },
+                        ResponsePayload::ProcStdout { id, line },
                     );
                     debug!(
                         "<Client @ {}> Sending response of type {}",
@@ -280,7 +330,7 @@ async fn proc_run(
                         break;
                     }
                 }
-                Ok(_) => break,
+                Ok(None) => break,
                 Err(_) => break,
             }
         }
@@ -289,16 +339,15 @@ async fn proc_run(
     // Spawn a task that sends stderr as a response
     let tx_2 = tx.clone();
     let tenant_2 = tenant.clone();
-    let mut stderr = child.stderr.take().unwrap();
+    let mut stderr = BufReader::new(child.stderr.take().unwrap()).lines();
     let stderr_task = tokio::spawn(async move {
         loop {
-            let mut data = Vec::new();
-            match stderr.read_to_end(&mut data).await {
-                Ok(n) if n > 0 => {
+            match stderr.next_line().await {
+                Ok(Some(line)) => {
                     let res = Response::new(
                         tenant_2.as_str(),
                         None,
-                        ResponsePayload::ProcStderr { id, data },
+                        ResponsePayload::ProcStderr { id, line },
                     );
                     debug!(
                         "<Client @ {}> Sending response of type {}",
@@ -309,7 +358,7 @@ async fn proc_run(
                         break;
                     }
                 }
-                Ok(_) => break,
+                Ok(None) => break,
                 Err(_) => break,
             }
         }
@@ -317,10 +366,20 @@ async fn proc_run(
 
     // Spawn a task that sends stdin to the process
     let mut stdin = child.stdin.take().unwrap();
-    let (stdin_tx, mut stdin_rx) = mpsc::channel::<Vec<u8>>(1);
+    let (stdin_tx, mut stdin_rx) = mpsc::channel::<String>(1);
     tokio::spawn(async move {
-        while let Some(data) = stdin_rx.recv().await {
-            if let Err(x) = stdin.write_all(&data).await {
+        while let Some(mut line) = stdin_rx.recv().await {
+            // NOTE: If the line coming in does not have a newline character,
+            //       we must add it to properly flush to stdin of process
+            //
+            //       Given that our data structure is called line, we assume
+            //       that it represents a complete line whether or not it
+            //       ends with the linefeed character
+            if !line.ends_with('\n') {
+                line.push('\n');
+            }
+
+            if let Err(x) = stdin.write_all(line.as_bytes()).await {
                 error!("Failed to send stdin to process {}: {}", id, x);
                 break;
             }
@@ -415,15 +474,7 @@ async fn proc_run(
         stdin_tx,
         kill_tx,
     };
-    state.lock().await.processes.insert(id, process);
-
-    state
-        .lock()
-        .await
-        .client_processes
-        .entry(addr)
-        .or_insert(Vec::new())
-        .push(id);
+    state.lock().await.push_process(addr, process);
 
     Ok(ResponsePayload::ProcStart { id })
 }
@@ -444,10 +495,10 @@ async fn proc_kill(state: HState, id: usize) -> Result<ResponsePayload, Box<dyn
 async fn proc_stdin(
     state: HState,
     id: usize,
-    data: Vec<u8>,
+    line: String,
 ) -> Result<ResponsePayload, Box<dyn Error>> {
     if let Some(process) = state.lock().await.processes.get(&id) {
-        process.stdin_tx.send(data).await.map_err(|_| {
+        process.stdin_tx.send(line).await.map_err(|_| {
             io::Error::new(io::ErrorKind::BrokenPipe, "Unable to send stdin to process")
         })?;
     }
